Module NLTE_Atom_structure
!
! Variable names are reminiscent of those used in Carlsson's MULTI:
! However, some MULTI features are not supported here for simplicity:
!    IW .ne. 0 is not supported
!    ITRAD .eq. 4 is not supported
  Integer, Parameter :: MaxNFreqs=100
  Integer, Parameter :: MaxNColStrings=500
!
  Type NLTE_Atom
!
! Variable definitions:
! element=2-character string with the element name (e.g., 'Ca')
! Z=Nuclear charge (atomic number)
! ion_stage=ionization stage of ground level (e.g., 2 for CaII)
! NK=Number of energy levels
! NLIN=Number of bound-bound transitions
! NCONT=Number of bound-free transitions in detail
! NFIX=Number of transitions treated with radiation temperatures
! Abund and AWgt=These values will not be read from the model atom file. 
!        They will be replaced with the contents of forward/atomic_data.f90
! 
     Character (Len=2) :: element
     Integer :: ion_stage, Z
     Integer :: NK, NLIN, NCNT, NFIX
     Real :: Abund, AWgt
!
! Use pointers for allocatable variables
!
! NK-sized vectors:
! cm_1=Energy level in cm^-1
! eV=Same in eV
! g=Statistical weight
! lvl_label=String with level label
! ion=Ionization stage the level belongs to
     Real, Dimension(:), pointer :: cm_1, eV, g
     Character (Len=256), Dimension(:), pointer :: lvl_label
     Integer, Dimension(:), pointer :: ion
!
! NLIN-sized vectors (bound-bound line transitions)
! Q0=Core-to-wing transition (in Doppler units, defined in NLTE_input)).
!       Wavelengths are spaced 
!       linearly in the core and logarithmically in the wings
! GA, GW, GS=Damping constants
! A=Einsteing A coefficient
     Real, Dimension(:), pointer :: Q0, GA, GW, GQ, A
!
! (NLIN+NCNT)-sized vectors (b-b and b-f radiative transitions):
! i=lower level index
! j=upper level index
! NQ=Number of pts for frequency quadrature
! QMAX=Furthest frequency (in Doppler units)
! f=Oscillator strength
! Alamb=Vacuum wavelength (in Angstroms)
     Integer, Dimension(:), pointer :: i, j, NQ
     Real, Dimension(:), pointer :: f, QMAX, Alamb
!
! NFIX-sized vectors (b-f fixed radiative transitions with radiation temp):
! iFIX=lower level index
! jFIX=upper level index
! ITrad=Photospheric-Chromospheric switch used in MULTI (can be 1, 2 or 3)
! IPho=1-continuum, 0-line
! A0=Cross-section at the limit (cm^-2)
! TRad=Radiation temperature (K)
! f=Oscillator strength
     Integer, Dimension(:), pointer :: iFIX, jFIX, ITRad, IPho
     Real, Dimension(:), pointer :: A0, TRad
! 
! Other arrays:
! Q(1:NQ(1:NLIN+NCNT),1:NLIN+NCNT): Wavelength grid (in Doppler units)
!          Note: If read from file ATOM via QMAX.lt.0 switch, then
!          units are in Angstroms, converted to Doppler in freqquad.f90
! FRQ(0:NQ(1:NLIN+NCNT),1:NLIN+NCNT): Freq grid for continuum trans (s^-1)
! WQ(1:NQ(1:NLIN+NCNT),1:NLIN+NCNT): Wavelength quadrature weights
! ALPHAC(1:NQ(1:NLIN+NCNT),1:NLIN+NCNT): Absorption profile
! WMU
! B(1:NK, 1:NK)=Einstein B coefficients (note the different dimensions wrt A)
! krad(1:NK,1:NK)=Index to transition between each pair of levels
     Real, Dimension(:,:), pointer :: Q, WQ, AlphaC, B, FRQ
     Integer, Dimension(:,:), pointer :: krad
!
! Information for collisional rates
! This is stored as strings that will be decoded by the corresponding routine
!
! ColRoutine=6-character array with the name of the routine to use.
!             Note: Currently, only CA2COL and GENCOL are implemented
! MaxNColStrings=Parameter setting the maximum number of lines with col data
! NColStrings=Actual of lines with collisional data
! ColStr(1:NColStrings)=Character*256 strings containing collisional data
!
     Character (Len=6) :: ColRoutine
     Integer :: NColStrings
     Character (Len=256), Dimension(MaxNColStrings) :: ColStr
  End Type NLTE_Atom
!
End Module NLTE_Atom_structure
