! This subroutine solves the scalar (unpolarized) radiative transfer
! equation using the Hermitian solution of Bellot Rubio, Ruiz Cobo and 
! Collados 1998, ApJ, 506, 805.
!
! The derivatives of the source function is computed assuming a 
! parabolic form.
!
Subroutine Scalar_hermite(npoints, tau_nu, S, Intensity)
  Implicit None
  Real, Dimension (npoints) :: tau_nu, S, Intensity
  Real, Dimension (npoints) :: S_p
  Real :: A, E, J, J_down, Cal_J, Cal_J_down, D, dtau
  Integer :: npoints, ibound, ipoint
  Real, Parameter :: Optically_thick = 100., Optically_thin=1.e-3
  Logical, Save :: Warning_thin=.TRUE., Warning_thick=.TRUE.
!
! Calculate derivatives of S
!
  Call Parab_deriv(npoints, tau_nu, S, S_p)
!
! Find the point for the boundary condition
!
  ibound=npoints-1
  Do while (tau_nu(ibound) .gt. Optically_thick .and. ibound .gt. 3)
     ibound=ibound-1
  End do
  If (tau_nu(ibound) .lt. 10. .and. Warning_thick) then
     Print *,'Warning! Line not optically thick at the bottom'
     Print *,'Tau_nu at the bottom = ',tau_nu(ibound)
     Print *,'This warning will not be shown any more during this run'
     Warning_thick=.FALSE.
  End if
!
! Set boundary condition (I=S + dS/d(tau_nu)) at ibound
!
  Intensity(ibound:npoints)=S(ibound:npoints)+S_p(ibound:npoints)
!
! Ray propagation
!
  ipoint=ibound-1
  Do while (ipoint .gt. 1 .and. tau_nu(ipoint) .gt. Optically_thin) ! Loop
     dtau=tau_nu(ipoint+1)-tau_nu(ipoint)
!
! Down quantities (point "i" in the paper)
!
     J_down=dtau/2.*S(ipoint+1)    
     Cal_J_down=dtau*dtau/12.*(S_p(ipoint+1) - S(ipoint+1))
!
! Up quantities (point "i+1" in the paper)
!
     J=dtau/2.*S(ipoint)
     Cal_J=dtau*dtau/12.*(S_p(ipoint) - S(ipoint))
     D=1. + dtau/2. + dtau*dtau/12.
     A=(1. - dtau/2. + dtau*dtau/12.) / D
     E=(J + J_down + Cal_J_down - Cal_J) / D
     Intensity(ipoint) = A*Intensity(ipoint+1) + E
     ipoint=ipoint-1
  End do
 ! End loop at the top of the atmo or where tau_nu < Optically_thin
  If (tau_nu(1) .gt. 1.e-2 .and. Warning_thin) then
     Print *,'Warning! Line not optically thin in the surface'
     Print *,'Tau_nu at the surface = ',tau_nu(1)
     Print *,'This warning will not be shown any more during this run'
     Warning_thin=.FALSE.
  End if
  Intensity(1:ipoint)=Intensity(ipoint+1)
  Return
End Subroutine Scalar_hermite
