module config_mod
use types_mod
use global_vars_mod
Use Param_structure
Use Model_structure
Use Line_data_structure
use maths_mod, only : hunt
implicit none
contains

!-------------------------------------------------------------------
! Initialize observations
!-------------------------------------------------------------------
	subroutine read_observations(observation)
	type(observations_type) :: observation
	character(len=128) :: obsfile
	integer :: i, ind, nlambda_new
	logical :: exists
	real(kind=8) :: obsmin, obsmax, neuralmin, neuralmax, minlambda, maxlambda, stepobs, stepneural
	real(kind=8), allocatable :: newlambda(:), newflux(:)
		
		inquire(file=trim(adjustl(observation%filename)), exist=exists)
		if (.not.exists) then
			write(*,*) 'File not found : ', trim(adjustl(observation%filename))
			stop
		endif
		
		open(unit=13,file=trim(adjustl(observation%filename)),action='read',status='old')
		read(13,*) observation%description
		read(13,*) observation%npoints
		if (verbose_level > 1) &
			write(*,*) 'Number of observed points : ', observation%npoints

		allocate(observation%obs_x(observation%npoints))
		allocate(observation%obs_y(4,observation%npoints))
		allocate(observation%obs_sigma(4,observation%npoints))
		allocate(observation%model_eval(4,observation%npoints))

		do i = 1, observation%npoints
			read(13,*) observation%obs_x(i), observation%obs_y(1,i), observation%obs_y(2,i), &
				observation%obs_y(3,i), observation%obs_y(4,i), observation%obs_sigma(1,i), &
				observation%obs_sigma(2,i), observation%obs_sigma(3,i), observation%obs_sigma(4,i)
		enddo

		close(13)

	end subroutine read_observations

!-------------------------------------------------------------------
! Set the index where the nodes are placed
!-------------------------------------------------------------------
	subroutine set_nodes(atmos)
	type(model_type) :: atmos
	integer :: i, j, k
	real(kind=8) :: x(10), xdistance
	integer :: pos
		
		do i = 1, atmos%ncomponents
			do j = 1, n_param_per_component-1
				select case(atmos%n_nodes(i,j))
					case(1)
! Set the node in the middle point
						x(1) = 0.5d0 * ( atmos%stratification(i)%ltau_500(1) + atmos%stratification(i)%ltau_500(Params%n_points) )
						call hunt(dble(atmos%stratification(i)%ltau_500), Params%n_points, x(1), pos)
						atmos%pos_nodes(i,j,1) = pos
					case(2)
! Set the nodes in the boundaries
						x(1) = atmos%stratification(i)%ltau_500(1)
						x(2) = atmos%stratification(i)%ltau_500(Params%n_points)						
						call hunt(dble(atmos%stratification(i)%ltau_500), Params%n_points, x(1), pos)
						atmos%pos_nodes(i,j,1) = pos
						call hunt(dble(atmos%stratification(i)%ltau_500), Params%n_points, x(2), pos)
						atmos%pos_nodes(i,j,2) = pos
					case(3:)
! Set the nodes equispaced in the log_tau axis
						xdistance = (atmos%stratification(i)%ltau_500(1) - atmos%stratification(i)%ltau_500(Params%n_points) ) / (atmos%n_nodes(i,j) - 1.d0)
						x(1) = atmos%stratification(i)%ltau_500(1)						
						call hunt(dble(atmos%stratification(i)%ltau_500), Params%n_points, x(1), pos)
						atmos%pos_nodes(i,j,1) = pos
						do k = 2, atmos%n_nodes(i,j)
							x(k) = x(k-1) - xdistance
							call hunt(dble(atmos%stratification(i)%ltau_500), Params%n_points, x(k), pos)
							atmos%pos_nodes(i,j,k) = pos
						enddo
				end select
			enddo
		enddo		
	end subroutine set_nodes

!-------------------------------------------------------------------
! Read file with atmospheric model
!-------------------------------------------------------------------
	subroutine read_model(filename, stratification)
	character(len=128) :: filename
	type (Model) :: stratification
	real(kind=8) :: t1, t2
	integer :: i
	logical :: exists

		inquire(file=trim(adjustl(filename)), exist=exists)
		if (.not.exists) then
			write(*,*) 'File not found : ', trim(adjustl(filename))
			stop
		endif

		open(unit=13, file=trim(adjustl(filename)), action='read',status='old')
		read(13,*)
		read(13,*)
		read(13,*) Params%n_points
		
		Call Allocate_model(stratification, Params%n_points)
		read(13,*) t1, t2
		stratification%v_mac = t1
		stratification%stray = t2
 		do i = Params%n_points, 1, -1
			read(13,*) stratification%ltau_500(i), stratification%temp(i), stratification%el_p(i), &
				stratification%v_mic(i), stratification%b_str(i), stratification%v_los(i), stratification%b_inc(i), &
				stratification%b_azi(i)
			stratification%v_los(i) = stratification%v_los(i) * 1.d-5
			stratification%v_mic(i) = stratification%v_mic(i) * 1.d-5
		enddo

		close(13)

		stratification%el_p(1) = 6.d-2 ! Boundary condition for electron pressure at the top
		stratification%expansion = 0.d0
	end subroutine read_model

!-------------------------------------------------------------------
! Initialize problem
!-------------------------------------------------------------------
	subroutine initialize_problem(inputfile, chain, prior)
	character(len=128) :: inputfile
	type(prior_type) :: prior
	type(markov_chain_type) :: chain
	integer :: nitermax, nparam, i, j
	logical :: exists

		inquire(file=trim(adjustl(inputfile)), exist=exists)
		if (.not.exists) then
			write(*,*) 'File not found : ', trim(adjustl(inputfile))
			stop
		endif

		open(unit=12,file=trim(adjustl(inputfile)),action='read',status='old')

!****************
! Some general inputs
!****************
! Number of lines and regions
		read(12,*) Params%n_lines
		read(12,*) Params%n_regions

! Number of components
		read(12,*) atmos%ncomponents

! Heliocentric angle
		read(12,*) Params%heliocentric

! File with observations
		read(12,*) obs%filename

! File with output sampling data
		read(12,*) chain%filename

! Read some information for Multinest
		read(12,*) nest_tol
		read(12,*) nest_efr
		read(12,*) nest_maxModes
		read(12,*) nest_nlive

		allocate(atmos%stratification(atmos%ncomponents))
		allocate(atmos%perturbed(atmos%ncomponents))
		allocate(atmos%filename(atmos%ncomponents))
		allocate(atmos%n_nodes(atmos%ncomponents,n_param_per_component))
		allocate(atmos%priors(atmos%ncomponents))
		allocate(atmos%n_free_parameters_per_component(atmos%ncomponents))
		allocate(atmos%ff(atmos%ncomponents))
		allocate(chain%param_names(chain%nparam))

!****************
! Read atmospheric models for all the components
!****************
		if (verbose_level > 1) &
			write(*,*) 'Reading atmospheric models...'
		do i = 1, atmos%ncomponents
			read(12,*) atmos%filename(i)
			call Read_model(atmos%filename(i), atmos%stratification(i))
			call Read_model(atmos%filename(i), atmos%perturbed(i))

! Read the number of nodes in each variable
			read(12,*) (atmos%n_nodes(i,j),j=1,n_param_per_component)

! Read the priors
			atmos%n_free_parameters_per_component(i) = sum(atmos%n_nodes(i,:))

			allocate(atmos%priors(i)%typ(atmos%n_free_parameters_per_component(i)))
			allocate(atmos%priors(i)%sigma(atmos%n_free_parameters_per_component(i)))
			allocate(atmos%priors(i)%mu(atmos%n_free_parameters_per_component(i)))
			allocate(atmos%priors(i)%minim(atmos%n_free_parameters_per_component(i)))
			allocate(atmos%priors(i)%maxim(atmos%n_free_parameters_per_component(i)))
			allocate(atmos%priors(i)%param_names(atmos%n_free_parameters_per_component(i)))

			do j = 1, n_param_per_component
				read(12,*) atmos%priors(i)%param_names(j)
				read(12,*) atmos%priors(i)%typ(j), atmos%priors(i)%minim(j), atmos%priors(i)%maxim(j), &
					atmos%priors(i)%mu(j), atmos%priors(i)%sigma(j)

! If Dirac prior, reduce the minimum and maximum to an interval around the mu value
				if (atmos%priors(i)%typ(j) == 2) then
					atmos%priors(i)%minim(j) = atmos%priors(i)%mu(j) - 1.d-3
 					atmos%priors(i)%maxim(j) = atmos%priors(i)%mu(j) + 1.d-3
 				endif
			enddo

		enddo

! Allocate memory for the position of the nodes
 		allocate(atmos%pos_nodes(atmos%ncomponents,n_param_per_component,maxval(atmos%n_nodes)))
 		allocate(atmos%value_nodes(atmos%ncomponents,n_param_per_component,maxval(atmos%n_nodes)))

		if (verbose_level > 1) &
			write(*,*) 'Total number of free parameters = ', sum(atmos%n_free_parameters_per_component)

		Hydro = .True. ! Set model in hydrostatic equilibrium

!****************
! Regions
!****************
		if (verbose_level > 1) &
			write(*,*) 'Reading region information...'
		allocate(Region(Params%n_regions))
		do i = 1, Params%n_regions
			read(12,*) Region(i)%nwavelengths, Region(i)%first_wlength, Region(i)%wave_step, Region(i)%Macro_enh, Region(i)%layer
		enddo

!****************
! Lines
!****************
		if (verbose_level > 1) &
			write(*,*) 'Reading line information...'
		allocate(Line(Params%n_lines))
		do i = 1, Params%n_lines
			read(12,*) Line(i)%Wlength, Line(i)%Energy_low, Line(i)%VDW_enh, Line(i)%loggf, Line(i)%Atomic_number, Line(i)%Ion_stage, &
				Line(i)%Atomic_weight
			read(12,*) Line(i)%Mult_low, Line(i)%Desig_low, Line(i)%J_low
			read(12,*) Line(i)%Mult_up, Line(i)%Desig_up, Line(i)%J_up
			Line(i)%collisions = 1
			Line(i)%Width = 1.0
			Line(i)%NLTEtransition = 0
			allocate(Line(i)%b_low(Params%n_points))
			allocate(Line(i)%b_up(Params%n_points))
			Line(i)%b_up = 1.0
			Line(i)%b_low = 1.0
		enddo

!****************
! Read observations
!****************
		if (verbose_level > 1) &
			write(*,*) 'Reading file with observations...'
		call read_observations(obs)
		Params%n_data = 4 * obs%npoints

!****************
! Other parameters
!****************
		Params%printout = 1 ! Verbosity level
		Params%speed = 2 ! Some speed optimizations
		Params%reference_cont = 3 ! How to normalize emergent intensity
		Params%Skip_lambda = 1 ! Leave it at 1
		Params%Update_opac = 1 ! How many Angstroms before recomputing background opac

		close(12)

!****************
! Set some information for the sampling
!****************
		chain%nparam = sum(atmos%n_free_parameters_per_component)
		chain%nparam_not_ffs = sum(atmos%n_nodes(:,1:n_param_per_component-1))
		chain%ncalls = 0
		allocate(chain%proposed(chain%nparam))
		allocate(chain%most_probable(chain%nparam))
		allocate(chain%best_parameters(chain%nparam))

		allocate(Stokes_Syn(Params%n_data))
		allocate(Stokes_Tot(Params%n_data))
		allocate(Params%Stray_prof(Params%n_data))
		Params%Stray_prof = 0.0

!****************
! Set the position of the nodes
!****************
 		call set_nodes(atmos)
		
	end subroutine initialize_problem

end module config_mod