! This subroutine solves the Stokes radiative transfer equation using the
! Hermitian solution of Bellot Rubio, Ruiz Cobo and Collados 1998, 
! ApJ, 506, 805.
! The derivatives of the source function and the absorption matrix are
! computed assuming a parabolic form for them.
!
Subroutine Hermite(npoints, ltau_500, Ab, S, Emergent_stokes)
  Implicit None
  Real, Dimension (npoints) :: ltau_500, S, S_p, Tmp, tau_nu, &
       tau_500
  Real, Dimension (4, npoints) :: Stokes
  Real, Dimension (4, 4, npoints) :: Ab, Ab_p
  Real, Dimension (4, 4) :: K2, A, Cal_K, D, Id, Tmp44
  Real, Dimension (4) :: Emergent_stokes, Tmp4
  Real, Dimension (4) :: J, Cal_J, E, J_down, Cal_J_down
  Real :: dtau
  Integer :: ipoint, npoints, ibound
  Real, Parameter :: Optically_thick = 100., Optically_thin=1.e-3
  Logical, Save :: Warning_thin=.TRUE., Warning_thick=.TRUE.
!
! First construct the tau_nu (monochromatic optical depth) and
! tau_5000 scales
!
  tau_500(1:npoints)=10**ltau_500(1:npoints)
  tau_nu(1)=(10**ltau_500(1))*Ab(1, 1, 1)
  Do ipoint=2, npoints
     tau_nu(ipoint)=tau_nu(ipoint-1)+ &
          (tau_500(ipoint)-tau_500(ipoint-1))* &
          0.5*(Ab(1, 1, ipoint)+Ab(1, 1, ipoint-1))
  End do
!
! Calculate the derivatives of Source funct. and Absorpt. mat.
!
  Call Parab_deriv(npoints, tau_500, Ab(1, 1, 1:npoints), Tmp)
  Ab_p(1, 1, 1:npoints)=Tmp
  Call Parab_deriv(npoints, tau_500, Ab(1, 2, 1:npoints), Tmp)
  Ab_p(1, 2, 1:npoints)=Tmp
  Call Parab_deriv(npoints, tau_500, Ab(1, 3, 1:npoints), Tmp)
  Ab_p(1, 3, 1:npoints)=Tmp
  Call Parab_deriv(npoints, tau_500, Ab(1, 4, 1:npoints), Tmp)
  Ab_p(1, 4, 1:npoints)=Tmp
  Call Parab_deriv(npoints, tau_500, Ab(2, 3, 1:npoints), Tmp)
  Ab_p(2, 3, 1:npoints)=Tmp
  Call Parab_deriv(npoints, tau_500, Ab(2, 4, 1:npoints), Tmp)
  Ab_p(2, 4, 1:npoints)=Tmp
  Call Parab_deriv(npoints, tau_500, Ab(3, 4, 1:npoints), Tmp)
  Ab_p(3, 4, 1:npoints)=Tmp
! Invoke symmetry properties to construct the rest of the abs. matrix derivat.
  Ab_p(2, 2, 1:npoints)=Ab_p(1, 1, 1:npoints)
  Ab_p(3, 3, 1:npoints)=Ab_p(1, 1, 1:npoints)
  Ab_p(4, 4, 1:npoints)=Ab_p(1, 1, 1:npoints)
  Ab_p(2, 1, 1:npoints)=Ab_p(1, 2, 1:npoints)
  Ab_p(3, 1, 1:npoints)=Ab_p(1, 3, 1:npoints)
  Ab_p(4, 1, 1:npoints)=Ab_p(1, 4, 1:npoints)
  Ab_p(3, 2, 1:npoints)=-Ab_p(2, 3, 1:npoints)
  Ab_p(4, 2, 1:npoints)=-Ab_p(2, 4, 1:npoints)
  Ab_p(4, 3, 1:npoints)=-Ab_p(3, 4, 1:npoints)
! Derivatives of source function (1st component only)
  Call Parab_deriv(npoints, tau_500, S, S_p)
!
! Find the point for the boundary condition
!
  ibound=npoints-1
  Do while (tau_nu(ibound) .gt. Optically_thick .and. ibound .gt. 3)
     ibound=ibound-1
  End do
  If (tau_nu(ibound) .lt. 10. .and. Warning_thick) then
     Print *,'Warning! Line not optically thick at the bottom'
     Print *,'Tau_nu at the bottom = ',tau_nu(ibound)
     Print *,'This warning will not be shown any more during this run'
     Warning_thick=.FALSE.
  End if
!
! Set boundary condition (I=S + dS/dtau) at ibound
!
  Stokes(1, ibound)=S(ibound) + S_p(ibound)/ & ! Boundary cond.
       Ab(1, 1, ibound) ! Note that S_p is with respect to
                        ! tau_500 and we need it with tau_nu
  Stokes(2:4, ibound)=0.
!
! Ray propagation
!
  Id(1:4, 1:4)=0.
  Id(1,1)=1.
  Id(2,2)=1.
  Id(3,3)=1.
  Id(4,4)=1. ! Define identity matrix
  ipoint=ibound-1
  Do while (ipoint .gt. 1 .and. tau_nu(ipoint) .gt. Optically_thin) ! Loop
     dtau=tau_500(ipoint+1)-tau_500(ipoint)
!
! Down quantities (point "i" in the paper)
!
     J_down(1:4)=dtau/2.*Ab(1:4, 1, ipoint+1)*S(ipoint+1)
     Call Multiply_matrix(4, 4, 4, Ab(1:4, 1:4, ipoint+1), &
          Ab(1:4, 1:4, ipoint+1), K2) ! K^2
     Cal_J_down(1:4)=dtau*dtau/12.*(Ab(1:4, 1, ipoint+1)*S_p(ipoint+1)+ &
          Ab_p(1:4, 1, ipoint+1)*S(ipoint+1) - K2(1:4, 1)*S(ipoint+1))
!
! Up quantities (point "i+1" in the paper)
!
     J(1:4)=dtau/2.*Ab(1:4, 1, ipoint)*S(ipoint)
     Call Multiply_matrix(4, 4, 4, Ab(1:4, 1:4, ipoint+1), &
          Ab(1:4, 1:4, ipoint), K2) ! K^2
     Cal_J(1:4)=dtau*dtau/12.*(Ab(1:4, 1, ipoint)*S_p(ipoint)+ &
          Ab_p(1:4, 1, ipoint)*S(ipoint) - K2(1:4, 1)*S(ipoint))
     Cal_K(1:4, 1:4)=K2(1:4, 1:4) - Ab_p(1:4, 1:4, ipoint)
     D(1:4, 1:4)=Id(1:4, 1:4) + dtau/2.*Ab(1:4, 1:4, ipoint) + &
          dtau*dtau/12.*Cal_K(1:4, 1:4)
     Call matinx(D) ! Now the matrix D is actually the D^-1 in the paper
     Tmp44(1:4, 1:4)=Id(1:4, 1:4) - dtau/2.*Ab(1:4, 1:4, ipoint+1) + &
          dtau*dtau/12.*Cal_K(1:4, 1:4)
     Call Multiply_matrix(4, 4, 4, D, Tmp44, A)
     Tmp4(1:4)=J(1:4)+J_down(1:4)+Cal_J_down(1:4)-Cal_J(1:4)
     Call Multiply_matrix(4, 4, 1, D, Tmp4, E)
     Call Multiply_matrix(4, 4, 1, A, Stokes(1:4, ipoint+1), Tmp4)
     Stokes(1:4, ipoint)=Tmp4(1:4) + E(1:4) ! I(i+1)=A.I(i)+E
     ipoint=ipoint-1
  End do ! End loop at the top of the atmo or where tau_nu < Optically_thin
  If (tau_nu(1) .gt. 1.e-2 .and. Warning_thin) then
     Print *,'Warning! Line not optically thin in the surface'
     Print *,'Tau_nu at the surface = ',tau_nu(1)
     Print *,'This warning will not be shown any more during this run'
     Warning_thin=.FALSE.
  End if
  Emergent_Stokes(1:4)=Stokes(1:4, ipoint+1)
  Return
End Subroutine Hermite
