Subroutine FormalSolution(NLTE, imu, X, S, RNu, P, LStMuNu)
! Given the opacity and source function, this routine solves the
! scalar radiative transfer equation. It returns P (the average
! incoming and outgoing intensity) and LStMuNu (the monochromatic local 
! Lambda-operator.
!
! X(1:NDEP)=Total opacity/Continuum opacity at 500nm
! S(1:NDEP)=Source function (arbitrary units)
! RNu(1:NDEP)=Ratio of line to total opacity (needed to compute LStMuNu)
! IncidentInt=Incident Intensity (same units as S)
! P(1:NDEP)=Output (same units as S)
! LStMuNu(1:NDEP)=Output (dimensionless)
!
  Use Param_structure
  Use NLTE_vars
  Implicit None
  Type (NLTE_variables) :: NLTE
  Real, Parameter :: OptThin=1e-2, OptThick=1000.
  Integer :: imu, idepth,itran, BoundUp, BoundLow
  Real, Dimension(0:NLTE%NDEP) :: Iminus, Iplus
  Real, Dimension(NLTE%NDEP) :: X, S, P, LStMuNu, Tau_500, Tau_Nu, Dtau_Nu,&
       RNu
  Real :: IncidentInt, CMu, T
!
  Iminus(:)=0.
  Iplus(:)=0.
  BoundUp=1
  BoundLow=NLTE%NDEP
!
! Compute monochromatic optical depths
  CMu=1./NLTE%XMu(imu)
  Tau_500(:)=10**(NLTE%Atmo%ltau_500)
  T=Tau_500(1)*X(1)*CMu
  Tau_Nu(1)=T
  Dtau_Nu(1)=T
  Dtau_Nu(2:NLTE%NDEP)=.5*(X(2:NLTE%NDEP)+X(1:NLTE%NDEP-1))* &
       (Tau_500(2:NLTE%NDEP)-Tau_500(1:NLTE%NDEP-1))*CMu
  Do idepth=2, NLTE%NDEP
     Tau_Nu(idepth)=Tau_Nu(idepth-1)+DTau_Nu(idepth)
     If (Tau_Nu(idepth) .gt. OptThin .and. Tau_Nu(idepth-1) .lt. OptThin) &
          BoundUp=idepth-1
     If (Tau_Nu(idepth) .gt. OptThick .and. Tau_Nu(idepth-1) .lt. OptThick) &
          BoundLow=idepth
  End do

!
  Call ShortCharacteristics ! Change this routine to other solver if needed
!
  Return
!
  contains
!
    Subroutine ShortCharacteristics
!
! FORMAL SOLVER USING SHORT CARACTERISTICS (SEE GORDON, OLSON, KUNASZ
! 1987). 
!
      Implicit None
      Integer :: step, K, KINIT, KEND
      Real :: T, Ex, Ex1, E0, E1, E2, D2, D3, D4, &
           ALF, BET, GAM, DELTAI, SDELTA, DELTAL
      Double Precision :: DTM, DTP, EXPDTP, EXPDTM
      Logical, Save :: Warning=.False.
!
      LStMuNu(1:NLTE%NDEP)=0.
!
      T=Tau_Nu(BoundUp)
      If (T .lt. OptThin) then ! Optically thin
         Ex1=T*(1.-T*(.5-T*(0.1666667-T*0.041666667)))
      Else
         if (T .lt. 10) then
            Ex1=1.-Exp(-T)
         Else
            Ex1=1.
         End if
      End if
      Ex=1.-Ex1
      Iminus(0)=0.
      Iminus(1:BoundUp)=S(1:BoundUp)*(T-.5*T*T)
      If (Iminus(1) .lt. 0) Iminus(1)=0.
! Bottom boundary condition: S+dS/dTau_Nu
      Iplus(BoundLow:NLTE%NDEP)=S(BoundLow:NLTE%NDEP)+ &
           (S(BoundLow:NLTE%NDEP)-S(BoundLow-1:NLTE%NDEP-1)) &
           /DTau_Nu(BoundLow:NLTE%NDEP)
      Iminus(BoundLow:NLTE%NDEP)=S(BoundLow:NLTE%NDEP)+ &
           (S(BoundLow:NLTE%NDEP)-S(BoundLow-1:NLTE%NDEP-1)) &
           /DTau_Nu(BoundLow:NLTE%NDEP)
! If all is optically thick
      If (BoundLow .le. 2) then
         Iplus(1:BoundLow)=S(1:BoundLow)+ &
              (S(2:BoundLow+1)-S(1:BoundLow))/DTau_Nu(1:BoundLow)
         Iminus(1:BoundLow)=S(1:BoundLow)+ &
              (S(2:BoundLow+1)-S(1:BoundLow))/DTau_Nu(1:BoundLow)
         LStMuNu(:)=1.
      Else ! Do the transfer
         Do step=1,-1,-2 ! FIRST INWARDS AND THEN OUTWARDS
            If (step .eq. 1) then ! IF INWARDS
               KINIT=BoundUp
               KEND=BoundLow
            Else                ! IF OUTWARDS
               KINIT=BoundLow
               KEND=BoundUp
            End if
            Do K=KINIT+step, KEND-step, step
               IF (step .eq. 1) then
                  DTM=DTau_Nu(K)
                  DTP=DTau_Nu(K+1)
               Else
                  DTM=DTau_Nu(K+1)
                  DTP=DTau_Nu(K)
               End if
               If (DTM .ge. OptThin) then
                  EXPDTM=Exp(-DTM)
                  E0=1.-EXPDTM
                  E1=DTM-E0
                  E2=DTM*DTM-2.*E1
               Else
                  EXPDTM=1.-DTM+(DTM*DTM)/2. ! TAYLOR EXPANSION
                  D2=DTM*DTM ! TO IMPROVE PRECISION AT SMALL OPTICAL DEPTHS
                  D3=DTM*D2
                  D4=DTM*D3
                  E0=DTM-(D2/2.)
                  E1=(D2/2.)-(D3/6.)
                  E2=(D3/3.)-(D4/12.)
               End if
               ALF=E0+(E2-(DTP+2*DTM)*E1)/(DTM*(DTM+DTP))
               BET=((DTM+DTP)*E1-E2)/DTM/DTP
               GAM=(E2-DTM*E1)/(DTP*(DTM+DTP))
               DELTAI=ALF*S(K-STEP)+BET*S(K)+GAM*S(K+STEP)
               SDELTA=1.-RNu(K)
               DELTAL=BET*SDELTA
               If (step .eq. 1) then
                  Iminus(K)=Iminus(K-1)*EXPDTM+DELTAI
! Valgrind complains about Iminus(2) here
                  IF (Iminus(K) .lt. 0 .or. DELTAL .gt. 1.1) then
! COMPUTE AGAIN USING LINEAR INTERPOLATION
                     ALF=E0-E1/DTM
                     BET=E1/DTM
                     DELTAI=ALF*S(K-step)+BET*S(K)
                     Iminus(K)=Iminus(K-1)*EXPDTM+DELTAI
                     GAM=0.
                     LStMuNu(K)=LStMuNu(K)+BET*SDELTA
                  Else
                     LStMuNu(K)=LStMuNu(K)+DELTAL
                  End if
               Else
                  Iplus(K)=Iplus(K+1)*EXPDTM+DELTAI
                  IF (Iplus(K) .lt. 0 .or. DELTAL .gt. 1.1) THEN 
! COMPUTE AGAIN USING LINEAR INTERPOLATION
                     ALF=E0-E1/DTM
                     BET=E1/DTM
                     DELTAI=ALF*S(K-STEP)+BET*S(K)
                     Iplus(K)=Iplus(K+1)*EXPDTM+DELTAI
                     GAM=0.
                     LStMuNu(K)=LStMuNu(K)+BET*SDELTA
                  Else
                     LStMuNu(K)=LStMuNu(K)+DELTAL
                  End if
               End if
            End do ! Do in K
! USE LINEAR INTERPOLATION FOR THE LAST POINT
            IF (DTP .ge. OptThin) then ! DELTA-TAU BETWEEN THE LAST TWO POINTS
               EXPDTP=Exp(-DTP)        ! IS DTP
               E0=1.-EXPDTP 
               E1=DTP-E0      
            Else
               EXPDTP=1.-DTP+(DTP*DTP)/2. ! TAYLOR EXPANSION 
               E0=DTP-DTP*DTP/2.          ! TO IMPROVE PRECISION AT SMALL
               E1=DTP*DTP/2.              ! OPTICAL DEPTHS
            End if
            K=KEND
            ALF=E0-E1/DTP
            BET=E1/DTP
            DELTAI=ALF*S(K-STEP)+BET*S(K)
            SDELTA=1.-RNu(K)
            LStMuNu(K)=LStMuNu(K)+BET*SDELTA
            IF (Step .eq. 1) then
               Iminus(K)=Iminus(K-1)*EXPDTP+DELTAI
            Else
               Iplus(K)=Iplus(K+1)*EXPDTP+DELTAI
            End if
         End do ! Do in step
      End if ! End check if all is optically thick
      Iplus(1:BoundUp)=Iplus(BoundUp)
      P(1:NLTE%NDEP)=(Iplus(1:NLTE%NDEP)+Iminus(1:NLTE%NDEP))/2.
      Iplus(0)=2.*(Ex*P(1)+S(1)*0.5*Ex1**2)
      LStMuNu(1:NLTE%NDEP)=LStMuNu(1:NLTE%NDEP)/2.
      LStMuNu(BoundLow:NLTE%NDEP)=1.
      LStMuNu(NLTE%NDEP)=.5
      If (MinVal(P) .lt. 0. .or. &
           MinVal(LStMuNu) .lt. 0. .or. &
           MaxVal(LStMuNu) .gt. 1.) then
         If (.not. Warning) then
            Print *,'WARNING: Negative I or LStar <0 or >1'
            Warning=.True.
         End if
         Do K=1, NLTE%NDEP
            If (P(K) .lt. 0) then
               P(K)=1.e-10
            endif
            If (LStMuNu(K) .lt. 0.) LStMuNu(K)=1.e-10
            If (LStMuNu(K) .gt. 1.) LStMuNu(K)=0.999
         End do
      End if
      Return
    End Subroutine ShortCharacteristics
End Subroutine FormalSolution
