! This subroutine takes a model and determines the electron pressure,
! gas pressure, density and the height scale by solving the hydrostatic
! equilibrium equation. The boundary condition for this equation is the
! electron pressure at the top of the atmosphere and is taken from the
! model Atmo.
!
Subroutine Hydrostatic(Params, Atmo, Error)
  Use Param_structure
  Use Model_structure
  Use Atomic_data
  Use Background_opacity_module, Only: Background_opacity
  Implicit None
  Type (Parameters) :: Params
  Type (Model) :: Atmo
  Type (Model), Save :: Saved_Atmo
  Integer :: ipoint, npoints, iatom, niter
  Integer, Dimension(1) :: imin
  Real, Dimension (Params%n_points) :: Kappa, Tau
  Real, Dimension (10) :: Pp
  Real, Parameter :: nu500 = cc/(5000.*1.e-8), Min_temp = 3500., &
       Mu=12.566370614 ! Vacuum permeability (G^2 cm^3/erg)=4*Pi, P_m=B^2/2/Mu
  Real :: dtau, temp, dif, n2P, Scat, chi_0, chi_e, eta
  Real :: Avmolweight, Asum, Wsum, Pg_Old
  Logical :: Warning, Error
  Logical, Save :: FirstTime=.True.
!
! Do we need to redo the calculation?
  If (FirstTime) then
     Allocate(Saved_Atmo%temp(Params%n_points))
     Allocate(Saved_Atmo%z_scale(Params%n_points))
     Allocate(Saved_Atmo%gas_p(Params%n_points))
     Allocate(Saved_Atmo%rho(Params%n_points))
     Allocate(Saved_Atmo%el_p(Params%n_points))
     Allocate(Saved_Atmo%ne(Params%n_points))
     Allocate(Saved_Atmo%nH(Params%n_points))
     Allocate(Saved_Atmo%nHplus(Params%n_points))
     Allocate(Saved_Atmo%nHminus(Params%n_points))
     Allocate(Saved_Atmo%nH2(Params%n_points))
  Else
     If (MaxVal(Abs(Atmo%temp-Saved_Atmo%temp)) .lt. 200) then
        Atmo%z_scale=Saved_Atmo%z_scale
        Atmo%gas_p=Saved_Atmo%gas_p
        Atmo%rho=Saved_Atmo%rho
        Atmo%el_p=Saved_Atmo%el_p
        Atmo%ne=Saved_Atmo%ne
        Atmo%nH=Saved_Atmo%nH
        Atmo%nHplus=Saved_Atmo%nHplus
        Atmo%nHminus=Saved_Atmo%nHminus
        Atmo%nH2=Saved_Atmo%nH2
        Return
     End if
  End if
!
  FirstTime=.False.
  npoints=Params%n_points
  Warning=.FALSE.
  Tau=10.**Atmo%ltau_500
!
! Calculate the average molecular weight of the particles
!
  Wsum=0.
  Asum=0.
  Do iatom=1, 92
     Wsum=Wsum+At_weight(iatom)*10**(At_abund(iatom)-12.)
     Asum=Asum+10**(At_abund(iatom)-12.)
  End do
!
! Boundary condition
!
  temp=Atmo%Temp(1)  
  If (temp .lt. Min_temp) then ! Check temperature so that gasc won't crash
     Warning=.TRUE.
     temp=Min_temp

! Andres: added the following that was not present in the original version of NICOLE
     Atmo%Temp(1) = temp
  End if
  Call Gasc(temp, Atmo%El_p(1), Atmo%Gas_p(1), Pp) ! Get gas pressure
  Atmo%nH(1)=Pp(1)/Pp(8)*Atmo%El_p(1)/bk/Atmo%Temp(1) ! Neutral H
  Atmo%nHplus(1)=Pp(6)/Pp(8)*Atmo%El_p(1)/bk/Atmo%Temp(1) ! Ionized H
  Atmo%nHminus(1)=Pp(10)/Pp(8)*Atmo%El_p(1)/bk/Atmo%Temp(1) ! Negative H ion
  Atmo%nH2(1)=Pp(7)/Pp(8)*Atmo%El_p(1)/bk/Atmo%Temp(1) ! Molecular H
  n2P=BK*Atmo%Temp(1)
  Kappa(1)=Background_opacity(Atmo%Temp(1), Atmo%El_p(1), Atmo%nH(1)*n2P, &
       Atmo%nHminus(1)*n2P, Atmo%nHplus(1)*n2P, Atmo%nH2(1)*n2P, 0., 5000., Scat)
  Avmolweight=Wsum/(Asum+Pp(8)) ! Average molecular weight.
  Atmo%Rho(1)=Atmo%Gas_p(1)*Avmolweight/Avog/bk/temp ! Gas density
  Kappa(1)=Kappa(1)/Atmo%Rho(1) ! Convert to cm^2/g
  
  Atmo%Z_scale(1)=0. ! Boundary condition for the height scale
  Do ipoint=2, npoints
     dtau=10.**Atmo%ltau_500(ipoint) - 10.**Atmo%ltau_500(ipoint-1)
     temp=Atmo%Temp(ipoint)
     If (temp .lt. Min_temp) then ! Check temperature so that gasc won't crash
        Atmo%Temp(ipoint)=Min_temp

! Andres: added the following to avoid too low temperatures that make the chemical equilibrium crash
        temp = Min_temp
        Warning=.TRUE.
     End if

     Atmo%El_p(ipoint)=Atmo%El_p(ipoint-1) ! Initialization
     Call Gasc(temp, Atmo%El_p(ipoint), Atmo%Gas_p(ipoint), Pp) ! Get gas pressure
     Atmo%nH(ipoint)=Pp(1)/Pp(8)*Atmo%El_p(ipoint)/bk/Atmo%Temp(ipoint) ! Neutral H
     Atmo%nHplus(ipoint)=Pp(6)/Pp(8)*Atmo%El_p(ipoint)/bk/Atmo%Temp(ipoint) ! Ionized H
     Atmo%nHminus(ipoint)=Pp(10)/Pp(8)*Atmo%El_p(ipoint)/bk/Atmo%Temp(ipoint) ! Negative H ion
     Atmo%nH2(ipoint)=Pp(7)/Pp(8)*Atmo%El_p(ipoint)/bk/Atmo%Temp(ipoint) ! Molecular H
     n2P=BK*Atmo%Temp(ipoint)
     Kappa(ipoint)=Background_opacity(Atmo%Temp(ipoint), Atmo%El_p(ipoint), Atmo%nH(ipoint)*n2P, &
       Atmo%nHminus(ipoint)*n2P, Atmo%nHplus(ipoint)*n2P, Atmo%nH2(ipoint)*n2P, 0., 5000., Scat)
     Kappa(ipoint)=Kappa(ipoint)/Atmo%Rho(ipoint-1) ! Convert to cm^2/g
     Atmo%Gas_p(ipoint)=Atmo%Gas_p(ipoint-1) + &
          Gravity*dtau/Kappa(ipoint-1)! + & ! Hydrostatic equilibrium equation.
!          (Atmo%B_str(ipoint-1)**2  - &
!           Atmo%B_str(ipoint)**2) /8./Pi ! Magnetic pressure term

! Iterate to obtain gas_p(ipoint)
     dif=1.e10
     niter=0
     Do while (niter .lt. 10 .and. dif .gt. 1e-3)
        niter=niter+1
        Pg_Old=Atmo%Gas_p(ipoint)
        If (Params%def_abund .eq. 0) then
           Call Pe_from_Pg(temp, Atmo%El_p(ipoint), &
                Atmo%Gas_p(ipoint), Warning) ! Evaluate El_p from Gas_p at this point
        Else
           Call Pe_from_T_Pg(temp, Atmo%Gas_P(ipoint), Atmo%El_P(ipoint))
        End if
        Call Gasc(temp, Atmo%El_p(ipoint), Atmo%Gas_p(ipoint), Pp) ! Get gas pressure
        Atmo%nH(ipoint)=Pp(1)/Pp(8)*Atmo%El_p(ipoint)/bk/Atmo%Temp(ipoint) ! Neutral H
        Atmo%nHplus(ipoint)=Pp(6)/Pp(8)*Atmo%El_p(ipoint)/bk/Atmo%Temp(ipoint) ! Ionized H
        Atmo%nHminus(ipoint)=Pp(10)/Pp(8)*Atmo%El_p(ipoint)/bk/Atmo%Temp(ipoint) ! Negative H ion
        Atmo%nH2(ipoint)=Pp(7)/Pp(8)*Atmo%El_p(ipoint)/bk/Atmo%Temp(ipoint) ! Molecular H
        Kappa(ipoint)=Background_opacity(Atmo%Temp(ipoint), Atmo%El_p(ipoint), Atmo%nH(ipoint)*n2P, &
             Atmo%nHminus(ipoint)*n2P, Atmo%nHplus(ipoint)*n2P, Atmo%nH2(ipoint)*n2P, 0., 5000., Scat)
        Atmo%Rho(ipoint)=Atmo%Gas_p(ipoint)*Avmolweight/Avog/bk/temp ! Density
        Kappa(ipoint)=Kappa(ipoint)/Atmo%Rho(ipoint) ! Convert to cm^2/g
        Atmo%Gas_p(ipoint)=Atmo%Gas_p(ipoint-1) + &
             Gravity*dtau*.5*(1./Kappa(ipoint-1)+1./Kappa(ipoint))
!            (Atmo%B_str(ipoint-1)**2  - &
!             Atmo%B_str(ipoint)**2) /8./Pi ! Magnetic pressure term
        dif=Abs(Pg_Old-Atmo%Gas_p(ipoint))/(Pg_old+Atmo%Gas_p(ipoint))
     End Do
     If (dif .gt. 1e-3) Warning=.TRUE.
     Call Gasc(temp, Atmo%El_p(ipoint), Atmo%Gas_p(ipoint), Pp) ! Get gas pressure
     Atmo%nH(ipoint)=Pp(1)/Pp(8)*Atmo%El_p(ipoint)/bk/Atmo%Temp(ipoint) ! Neutral H
     Atmo%nHplus(ipoint)=Pp(6)/Pp(8)*Atmo%El_p(ipoint)/bk/Atmo%Temp(ipoint) ! Ionized H
     Atmo%nHminus(ipoint)=Pp(10)/Pp(8)*Atmo%El_p(ipoint)/bk/Atmo%Temp(ipoint) ! Negative H ion
     Atmo%nH2(ipoint)=Pp(7)/Pp(8)*Atmo%El_p(ipoint)/bk/Atmo%Temp(ipoint) ! Molecular H
     Avmolweight=Wsum/(Asum+Pp(8)) ! Average molecular weight.
     Atmo%Rho(ipoint)=Atmo%Gas_p(ipoint)*Avmolweight/Avog/bk/temp ! Gas density
     Atmo%Z_scale(ipoint)=Atmo%Z_scale(ipoint-1) - &
          dtau/2./1.e5* &
          (1./(Kappa(ipoint)*Atmo%Rho(ipoint))+1./(Kappa(ipoint-1)*Atmo%Rho(ipoint-1)))
  End do
  If (Warning) then 
!       print *,'Inaccuracy in hydrostatic equation'
     Error=.True.
  End if
  imin=MinLoc(Abs(Atmo%ltau_500))
  Atmo%Z_scale(1:Params%n_points)=Atmo%Z_scale(1:Params%n_points) - &
       Atmo%Z_scale(imin(1))
  Atmo%ne(1:Params%n_points)=Atmo%el_p(1:Params%n_points)/ &
       bk/Atmo%Temp(1:Params%n_points)
!
  Saved_Atmo%temp=Atmo%temp
  Saved_Atmo%z_scale=Atmo%z_scale
  Saved_Atmo%gas_p=Atmo%gas_p
  Saved_Atmo%rho=Atmo%rho
  Saved_Atmo%el_p=Atmo%el_p
  Saved_Atmo%ne=Atmo%ne
  Saved_Atmo%nH=Atmo%nH
  Saved_Atmo%nHplus=Atmo%nHplus
  Saved_Atmo%nHminus=Atmo%nHminus
  Saved_Atmo%nH2=Atmo%nH2
  Return
End Subroutine Hydrostatic
!
! Pe_from_Pg evaluates the electron pressure from the temperature, the
! gas pressure and an estimate of the electron pressure.
!
Subroutine Pe_from_Pg(T, Pe, Pg, Warning)
  Implicit None
  Real :: T, Pe, Pg, Pe_old, dif
  Real, Dimension (10) :: Pp
  Real, Parameter :: Precision = 0.1
  Integer, Parameter :: niters=25
  Integer :: iters
  Logical :: Warning
!
  Warning=.False.
  iters=0
  dif=1.
  Do while (dif .gt. Precision .and. iters .le. niters)
     Pe_old=Pe
     Call Gasc_inv(T, Pg, Pe)
     Pe=.5*(Pe+Pe_old) ! Smooth the variation
     dif=Abs(Pe-Pe_old)/(Pe+Pe_old)
     iters=iters+1
  End do
  If (dif .gt. Precision) then
     Warning=.TRUE.
  End if
  Return
End Subroutine Pe_from_Pg

