module maths_mod
use global_vars_mod, only : PI, atmos, n_param_per_component, Params
implicit none
contains
!-------------------------------------------------------------------
! Read n lines at the selected unit
!-------------------------------------------------------------------
	subroutine lb(u, n)
	integer :: u, n
   integer :: i
   	do i = 1, n
      	read(u,*)
      enddo
	end subroutine lb
        
!-------------------------------------------------------------
! Carry out the Cholesky decomposition of a symmetric matrix
!-------------------------------------------------------------
	subroutine cholesky(a,n,p)
	integer :: n
	real(kind=8) :: a(n,n), p(n)
	integer :: i, j, k
	real(kind=8) :: sum
	
		do i = 1, n
			do j = i, n
				sum = a(i,j)
				do k = i-1, 1, -1
					sum = sum - a(i,k)*a(j,k)
				enddo
				if (i == j) then
					if (sum == 0.d0) then
						print *, 'Cholesky decomposition failed...'	
					endif
					p(i) = sqrt(sum)
				else
					a(j,i) = sum / p(i)
				endif
			enddo
		enddo
						
	end subroutine cholesky
	
!-------------------------------------------------------------
! Initialize the random number generator
!-------------------------------------------------------------
	subroutine init_random_seed()
	integer :: i, n, clock
   integer, dimension(:), allocatable :: seed
          
   	call random_seed(size = n)
      allocate(seed(n))
          
      call system_clock(count=clock)
          
      seed = clock + 37 * (/ (i - 1, i = 1, n) /)
      call random_seed(put = seed)
          
      deallocate(seed)
   end subroutine init_random_seed
          
!-------------------------------------------------------------
! Generates a random number following an uniform distribution in the interval [0,1]
! Call it with idum<0 for initialization
!-------------------------------------------------------------
	function randomu()
	real(kind=8) :: randomu
	
		call random_number(randomu)
							
	end function randomu
		
!-------------------------------------------------------------
! Generates a random number following an normal distribution with zero mean
! and unit variance
!-------------------------------------------------------------
	function randomn()

	real(kind=8) :: randomn

	real(kind=8) :: u, sum
	real(kind=8), save :: v, sln
	logical, save :: second = .false.
	real(kind=8), parameter :: one = 1.0, vsmall = tiny( one )

! if second, use the second random number generated on last call
	if (second) then

		second = .false.
  		randomn = v*sln
	else
! first call; generate a pair of random normals

  		second = .true.
  		do
    		call random_number( u )
    		call random_number( v )
    		u = scale( u, 1 ) - one
    		v = scale( v, 1 ) - one
    		sum = u*u + v*v + vsmall         ! vsmall added to prevent log(zero) / zero
    		if(sum < one) exit
  		end do
  		sln = sqrt(- scale( log(sum), 1 ) / sum)
  		randomn = u*sln
	end if

	return
	end function randomn

! 	function randomn()
! 	real(kind=8) :: randomn, gasdev, fac, gset, rsq, v1, v2, ran1
! 	integer :: iset
! 	save :: iset, gset		
! 		if (iset == 0) then
! 			rsq = 0.0
! 			do while(rsq > 1 .or. rsq == 0.0)
! 				v1 = 2.0*randomu()-1.0
! 				v2 = 2.0*randomu()-1.0
! 				rsq = v1**2+v2**2
! 			enddo
! 			fac = sqrt(-2.0*log(rsq)/rsq)
! 			gset = v1*fac
! 			randomn = v2*fac
! 			iset = 1
! 		else
! 			randomn = gset
! 			iset = 0
! 		endif
! 	end function randomn
	
!-------------------------------------------------------------
! Generates a multivariate normal random number with a given
! mean and covariance matrix
!-------------------------------------------------------------
	function mrandomn(rmean,covar)
	integer :: idum, n, i, j
	real(kind=8) :: rmean(:), covar(:,:), mrandomn(size(rmean))
	real(kind=8) :: chol(size(rmean),size(rmean)), p(size(rmean)), eps(size(rmean))
	
		n = size(rmean)
				
		chol = covar
		
		do i = 1, n
			eps(i) = randomn()
			chol(i,i) = chol(i,i) + 1.d-7    ! Some regularization
		enddo
								
		call cholesky(chol,n,p)
										
		do j = 1, n			
			do i = j, n
				chol(j,i) = 0.d0
			enddo
			chol(j,j) = p(j)
		enddo
								
		mrandomn = matmul(chol,eps) + rmean			
				
		
	end function mrandomn
	
!-------------------------------------------------------------
! Return the diagonal of a matrix
!-------------------------------------------------------------
	function diagon(mat)
	real(kind=8) :: mat(:,:)
	real(kind=8) :: diagon(size(mat(1,:)))
	integer :: i
	
		do i = 1, size(mat(1,:))
			diagon(i) = mat(i,i)
		enddo
		
	end function diagon

!-------------------------------------------------------------
! CUBINT approximates an integral using cubic interpolation of data.
!  Parameters:
!
!    Input, real FTAB(NTAB), contains the tabulated function
!    values, FTAB(I) = F(XTAB(I)).
!
!    Input, real XTAB(NTAB), contains the points at which the
!    function was tabulated.  XTAB should contain distinct
!    values, given in ascending order.
!
!    Input, integer NTAB, the number of tabulated points.
!    NTAB must be at least 4.
!
!    Input, integer IA, the entry of XTAB at which integration
!    is to begin.  IA must be no less than 1 and no greater
!    than NTAB.
!
!    Input, integer IB, the entry of XTAB at which integration
!    is to end.  IB must be no less than 1 and no greater than
!    NTAB.
!
!    Output, real RESULT, the approximate value of the
!    integral from XTAB(IA) to XTAB(IB) of the function.
!
!    Output, real ERROR, an estimate of the error in
!    integration.
!-------------------------------------------------------------
	subroutine cubint ( ftab, xtab, ntab, ia, ib, result, error )

  	integer ntab
!
  	real(kind=8) :: c, d1, d2, d3, error, ftab(ntab), h1, h2, h3, h4
  	integer :: i, ia, ib, ind, it, j, k
  	real(kind=8) r1, r2, r3, r4, result, s, term, xtab(ntab)
!
  	result = 0.0E+00
  	error = 0.0E+00
 
  	if ( ia == ib ) then
    	return
  	end if
 
  	if ( ntab < 4 ) then
    	write ( *, '(a)' ) ' '
    	write ( *, '(a)' ) 'CUBINT - Fatal error!'
    	write ( *, '(a,i6)' ) '  NTAB must be at least 4, but input NTAB = ',ntab
    	stop
  	endif
 
  	if ( ia < 1 ) then
    	write ( *, '(a)' ) ' '
    	write ( *, '(a)' ) 'CUBINT - Fatal error!'
    	write ( *, '(a,i6)' ) '  IA must be at least 1, but input IA = ',ia
    	stop
  	endif
 
  	if ( ia > ntab ) then
   	write ( *, '(a)' ) ' '
    	write ( *, '(a)' ) 'CUBINT - Fatal error!'
    	write ( *, '(a,i6)' ) '  IA must be <= NTAB, but input IA=',ia
    	stop
  	endif
 
  	if ( ib < 1 ) then
    	write ( *, '(a)' ) ' '
    	write ( *, '(a)' ) 'CUBINT - Fatal error!'
    	write ( *, '(a,i6)' ) '  IB must be at least 1, but input IB = ',ib
    	stop
  	endif
 
  	if ( ib > ntab ) then
    	write ( *, '(a)' ) ' '
    	write ( *, '(a)' ) 'CUBINT - Fatal error!'
    	write ( *, '(a,i6)' ) '  IB must be <= NTAB, but input IB=',ib
    	stop
  	endif
!
!  Temporarily switch IA and IB, and store minus sign in IND
!  so that, while integration is carried out from low X's
!  to high ones, the sense of the integral is preserved.
!
  	if ( ia > ib ) then
    	ind = -1
    	it = ib
    	ib = ia
    	ia = it
  	else
    	ind = 1
  	endif
 
  	s = 0.0E+00
  	c = 0.0E+00
  	r4 = 0.0E+00
  	j = ntab-2
  	if ( ia < ntab-1 .or. ntab == 4 ) then
    	j=max(3,ia)
  	endif

  	k = 4
  	if ( ib > 2 .or. ntab == 4 ) then
    	k=min(ntab,ib+2)-1
  	endif
 
  	do i = j, k
 
    	if ( i <= j ) then
 
      	h2 = xtab(j-1)-xtab(j-2)
      	d3 = (ftab(j-1)-ftab(j-2)) / h2
      	h3 = xtab(j)-xtab(j-1)
      	d1 = (ftab(j)-ftab(j-1)) / h3
      	h1 = h2+h3
      	d2 = (d1-d3)/h1
      	h4 = xtab(j+1)-xtab(j)
      	r1 = (ftab(j+1)-ftab(j)) / h4
      	r2 = (r1-d1) / (h4+h3)
      	h1 = h1+h4
      	r3 = (r2-d2) / h1
 
      	if ( ia <= 1 ) then
        		result = h2 * (ftab(1)+h2*(0.5*d3-h2*(d2/6.0-(h2+h3+h3)*r3/12.)))
        		s = -h2**3 * (h2*(3.0*h2+5.0*h4)+10.0*h3*h1)/60.0
      	endif
 
    	else
 
	      h4 = xtab(i+1)-xtab(i)
      	r1 = (ftab(i+1)-ftab(i))/h4
      	r4 = h4+h3
      	r2 = (r1-d1)/r4
      	r4 = r4+h2
      	r3 = (r2-d2)/r4
      	r4 = (r3-d3)/(r4+h1)
 
    	endif
 
    	if ( i > ia .and. i <= ib ) then
 
      	term = h3*((ftab(i)+ftab(i-1))*0.5-h3*h3*(d2+r2+(h2-h4)*r3) / 12.0 )
      	result = result+term
      	c = h3**3*(2.0E+00 *h3*h3+5.*(h3*(h4+h2) + 2.0 * h2 * h4 ) ) / 120.0E+00
      	error = error+(c+s)*r4
 
      	if ( i /= j ) then
        		s = c
      	else
        		s = s+c+c
      	endif
 
    	else
 
	      error = error+r4*s
 
    	endif
 
    	if ( i >= k ) then
 
      	if ( ib >= ntab ) then
	        	term = h4*(ftab(ntab) - h4*(0.5*r1+h4*(r2/6.0 +(h3+h3+h4)*r3/12.)))
        		result = result + term
        		error = error - h4**3 * r4 * &
          		( h4 * ( 3.0 * h4 + 5.0 * h2 ) &
          		+ 10.0 * h3 * ( h2 + h3 + h4 ) ) / 60.0E+00
      	endif
 
      	if ( ib >= ntab-1 ) error=error+s*r4
    	else
	      h1 = h2
      	h2 = h3
      	h3 = h4
      	d1 = r1
      	d2 = r2
      	d3 = r3
    	endif
 
  	enddo
!
!  Restore original values of IA and IB, reverse signs
!  of RESULT and ERROR, to account for integration
!  that proceeded from high X to low X.
!
  	if ( ind /= 1 ) then
    	it = ib
    	ib = ia
    	ia = it
    	result = -result
    	error = -error
  	endif
 
  	return
	end subroutine 

!----------------------------------------------------------------
! This function integrates a tabulated function
!----------------------------------------------------------------		
	function int_tabulated(x, f)
	real(kind=8) :: x(:), f(:), int_tabulated, res, error_res
	integer :: n
		n = size(x)
		call cubint (f, x, n, 1, n, res, error_res)
		int_tabulated = res
	end function int_tabulated

	subroutine qsortd(x,ind,n)
	
	implicit none
	integer, parameter  :: dp = 4
	
	real (kind=8), intent(in)  :: x(:)
	integer, intent(out)   :: ind(:)
	integer, intent(in)    :: n
	
	!***************************************************************************
	
	!                                                         robert renka
	!                                                 oak ridge natl. lab.
	
	!   this subroutine uses an order n*log(n) quick sort to sort a real (dp)
	! array x into increasing order.  the algorithm is as follows.  ind is
	! initialized to the ordered sequence of indices 1,...,n, and all interchanges
	! are applied to ind.  x is divided into two portions by picking a central
	! element t.  the first and last elements are compared with t, and
	! interchanges are applied as necessary so that the three values are in
	! ascending order.  interchanges are then applied so that all elements
	! greater than t are in the upper portion of the array and all elements
	! less than t are in the lower portion.  the upper and lower indices of one
	! of the portions are saved in local arrays, and the process is repeated
	! iteratively on the other portion.  when a portion is completely sorted,
	! the process begins again by retrieving the indices bounding another
	! unsorted portion.
	
	! input parameters -   n - length of the array x.
	
	!                      x - vector of length n to be sorted.
	
	!                    ind - vector of length >= n.
	
	! n and x are not altered by this routine.
	
	! output parameter - ind - sequence of indices 1,...,n permuted in the same
	!                          fashion as x would be.  thus, the ordering on
	!                          x is defined by y(i) = x(ind(i)).
	
	!*********************************************************************
	
	! note -- iu and il must be dimensioned >= log(n) where log has base 2.
	
	!*********************************************************************
	
	integer   :: iu(21), il(21)
	integer   :: m, i, j, k, l, ij, it, itt, indx
	real(kind=8)      :: r
	real(kind=8) :: t
	
	! local parameters -
	
	! iu,il =  temporary storage for the upper and lower
	!            indices of portions of the array x
	! m =      index for iu and il
	! i,j =    lower and upper indices of a portion of x
	! k,l =    indices in the range i,...,j
	! ij =     randomly chosen index between i and j
	! it,itt = temporary storage for interchanges in ind
	! indx =   temporary index for x
	! r =      pseudo random number for generating ij
	! t =      central element of x
	
	if (n <= 0) return
	
	! initialize ind, m, i, j, and r
	
	do  i = 1, n
	ind(i) = i
	end do
	
	m = 1
	i = 1
	j = n
	r = .375
	
	! top of loop
	
	20 if (i >= j) go to 70
	if (r <= .5898437) then
	r = r + .0390625
	else
	r = r - .21875
	end if
	
	! initialize k
	
	30 k = i
	
	! select a central element of x and save it in t
	
	ij = i + r*(j-i)
	it = ind(ij)
	t = x(it)
	
	! if the first element of the array is greater than t,
	!   interchange it with t
	
	indx = ind(i)
	if (x(indx) > t) then
	ind(ij) = indx
	ind(i) = it
	it = indx
	t = x(it)
	end if
	
	! initialize l
	
	l = j
	
	! if the last element of the array is less than t,
	!   interchange it with t
	
	indx = ind(j)
	if (x(indx) >= t) go to 50
	ind(ij) = indx
	ind(j) = it
	it = indx
	t = x(it)
	
	! if the first element of the array is greater than t,
	!   interchange it with t
	
	indx = ind(i)
	if (x(indx) <= t) go to 50
	ind(ij) = indx
	ind(i) = it
	it = indx
	t = x(it)
	go to 50
	
	! interchange elements k and l
	
	40 itt = ind(l)
	ind(l) = ind(k)
	ind(k) = itt
	
	! find an element in the upper part of the array which is
	!   not larger than t
	
	50 l = l - 1
	indx = ind(l)
	if (x(indx) > t) go to 50
	
	! find an element in the lower part of the array whcih is not smaller than t
	
	60 k = k + 1
	indx = ind(k)
	if (x(indx) < t) go to 60
	
	! if k <= l, interchange elements k and l
	
	if (k <= l) go to 40
	
	! save the upper and lower subscripts of the portion of the
	!   array yet to be sorted
	
	if (l-i > j-k) then
	il(m) = i
	iu(m) = l
	i = k
	m = m + 1
	go to 80
	end if
	
	il(m) = k
	iu(m) = j
	j = l
	m = m + 1
	go to 80
	
	! begin again on another unsorted portion of the array
	
	70 m = m - 1
	if (m == 0) return
	i = il(m)
	j = iu(m)
	
	80 if (j-i >= 11) go to 30
	if (i == 1) go to 20
	i = i - 1
	
	! sort elements i+1,...,j.  note that 1 <= i < j and j-i < 11.
	
	90 i = i + 1
	if (i == j) go to 70
	indx = ind(i+1)
	t = x(indx)
	it = indx
	indx = ind(i)
	if (x(indx) <= t) go to 90
	k = i
	
	100 ind(k+1) = ind(k)
	k = k - 1
	indx = ind(k)
	if (t < x(indx)) go to 100
	
	ind(k+1) = it
	go to 90
	end subroutine qsortd

! ---------------------------------------------------------
! Given x(:) and y(:) which tabulate a function and the derivative at the boundary points
! this function returns the second derivative of the spline at each point
! ---------------------------------------------------------
		subroutine splin1(x,y,yp1,ypn,y2)
		real(kind=8), INTENT(IN) :: x(:), y(:), yp1, ypn
		real(kind=8), INTENT(INOUT) :: y2(size(x))
		integer :: n, i, k
		real(kind=8) :: p, qn, sig, un, u(size(x))

			n = size(x)
			
			if (yp1 > .99e30) then
				y2(1) = 0.e0
				u(1) = 0.e0
			else
				y2(1) = -0.5e0
				u(1) = (3.e0/(x(2)-x(1)))*((y(2)-y(1))/(x(2)-x(1))-yp1)
			endif

			do i = 2, n-1
				sig = (x(i)-x(i-1))/(x(i+1)-x(i-1))				
				p = sig * y2(i-1)+2.e0
				y2(i) = (sig-1.e0)/p
				u(i) = (6.e0*((y(i+1)-y(i))/(x(i+1)-x(i))-(y(i)-y(i-1))/(x(i)-x(i-1)))/&
					(x(i+1)-x(i-1))-sig*u(i-1))/p
			enddo
			if (ypn > .99e30) then
				qn = 0.e0
				un = 0.e0
			else
				qn = 0.5e0
				un = (3.e0/(x(n)-x(n-1)))*(ypn-(y(n)-y(n-1))/(x(n)-x(n-1)))
			endif
			
			y2(n) = (un-qn*u(n-1))/(qn*y2(n-1)+1.e0)

			do k = n-1, 1, -1
				y2(k) = y2(k)*y2(k+1)+u(k)
			enddo

		end subroutine splin1

! ---------------------------------------------------------
! Given xa(:) and ya(:) which tabulate a function, returns the interpolation using
! splines of vector x(:) in y(:)
! ---------------------------------------------------------
		subroutine spline(xa,ya,x,y)
		real(kind=8), INTENT(INOUT) :: y(:)
		real(kind=8), INTENT(IN) :: xa(:), ya(:), x(:)
		real(kind=8) :: y2a(size(xa))
		integer :: n_x, n, i, k, khi, klo
		real(kind=8) :: a, b, h, extrap
			
			n = size(xa)
			n_x = size(x)
			call splin1(xa,ya,1.d30,1.d30,y2a)

			do i = 1, n_x					

! Downward extrapolation 
				if (x(i) < xa(1)) then
!					y(i) = ya(1)
					y(i) = ya(1) + (ya(1)-ya(2))/(xa(1)-xa(2)) * (xa(1) - x(i))
				else 

! Upward extrapolation
				if (x(i) > xa(n)) then
!					y(i) = ya(n)
					y(i) = ya(n) + (ya(n)-ya(n-1)) / (xa(n)-xa(n-1)) * (x(i) - xa(n))
				else
! In range
						klo = 1
						khi = n
1						if(khi-klo > 1) then
							k = (khi+klo)/2
							if (xa(k) > x(i)) then
								khi = k
							else
								klo = k
							endif					
							go to 1
						endif

						h = xa(khi)-xa(klo)

						if (h == 0.d0) pause 'bad xa input in spline'
						a = (xa(khi)-x(i))/h
						b = (x(i)-xa(klo))/h

						y(i) = a*ya(klo)+b*ya(khi)+((a**3.e0-a)*y2a(klo)+(b**3.e0-b)*y2a(khi))*(h**2.e0)/6.e0
					endif
				endif
			enddo

		end subroutine spline

! ---------------------------------------------------------
! Calculates the median of a given vector
! n.b. on exit, the array x is partially ordered.
! ---------------------------------------------------------
	subroutine median(x, n, xmed)


		integer, intent(in)                :: n
		real(kind=8), intent(in out), dimension(:) :: x
		real(kind=8), intent(out)                  :: xmed
		
		! local variables
		
		real(kind=8)    :: temp, xhi, xlo, xmax, xmin
		logical :: odd
		integer :: hi, lo, nby2, nby2p1, mid, i, j, k
		
		nby2 = n / 2
		nby2p1 = nby2 + 1
		odd = .true.
		
		!     hi & lo are position limits encompassing the median.
		
		if (n == 2 * nby2) odd = .false.
		lo = 1
		hi = n
		if (n < 3) then
		if (n < 1) then
			xmed = 0.0
			return
		end if
		xmed = x(1)
		if (n == 1) return
		xmed = 0.5*(xmed + x(2))
		return
		end if
		
		!     find median of 1st, middle & last values.
		
		10 mid = (lo + hi)/2
		xmed = x(mid)
		xlo = x(lo)
		xhi = x(hi)
		if (xhi < xlo) then          ! swap xhi & xlo
		temp = xhi
		xhi = xlo
		xlo = temp
		end if
		if (xmed > xhi) then
		xmed = xhi
		else if (xmed < xlo) then
		xmed = xlo
		end if
		
		! the basic quicksort algorithm to move all values <= the sort key (xmed)
		! to the left-hand end, and all higher values to the other end.
		
		i = lo
		j = hi
		50 do
		if (x(i) >= xmed) exit
		i = i + 1
		end do
		do
		if (x(j) <= xmed) exit
		j = j - 1
		end do
		if (i < j) then
		temp = x(i)
		x(i) = x(j)
		x(j) = temp
		i = i + 1
		j = j - 1
		
		!     decide which half the median is in.
		
		if (i <= j) go to 50
		end if
		
		if (.not. odd) then
		if (j == nby2 .and. i == nby2p1) go to 130
		if (j < nby2) lo = i
		if (i > nby2p1) hi = j
		if (i /= j) go to 100
		if (i == nby2) lo = nby2
		if (j == nby2p1) hi = nby2p1
		else
		if (j < nby2p1) lo = i
		if (i > nby2p1) hi = j
		if (i /= j) go to 100
		
		! test whether median has been isolated.
		
		if (i == nby2p1) return
		end if
		100 if (lo < hi - 1) go to 10
		
		if (.not. odd) then
		xmed = 0.5*(x(nby2) + x(nby2p1))
		return
		end if
		temp = x(lo)
		if (temp > x(hi)) then
		x(lo) = x(hi)
		x(hi) = temp
		end if
		xmed = x(nby2p1)
		return
		
		! special case, n even, j = n/2 & i = j + 1, so the median is
		! between the two halves of the series.   find max. of the first
		! half & min. of the second half, then average.
		
		130 xmax = x(1)
		do k = lo, j
		xmax = max(xmax, x(k))
		end do
		xmin = x(n)
		do k = i, hi
		xmin = min(xmin, x(k))
		end do
		xmed = 0.5*(xmin + xmax)

		return
	end subroutine median

! ---------------------------------------------------------
!	Given an array xx(1:n), and given a value x, returns a value jlo such that x is between
!	xx(jlo) and xx(jlo+1). xx(1:n) must be monotonic, either increasing or decreasing.
!	jlo=0 or jlo=n is returned to indicate that x is out of range. jlo on input is taken as
!	the initial guess for jlo on output.
! ---------------------------------------------------------
	subroutine hunt(xx,n,x,jlo)
	integer :: jlo,n
	real(kind=8) :: x,xx(n)
	integer :: inc,jhi,jm
	logical :: ascnd

		ascnd=xx(n).ge.xx(1)
		if (jlo.le.0.or.jlo.gt.n) then
			jlo=0
			jhi=n+1
			goto 3
		endif
		inc=1
		if (x.ge.xx(jlo).eqv.ascnd) then
1     	jhi=jlo+inc
			if (jhi.gt.n) then
				jhi=n+1
			else if (x.ge.xx(jhi).eqv.ascnd) then
				jlo=jhi
				inc=inc+inc
				goto 1
			endif
		else
			jhi=jlo
2     	jlo=jhi-inc
			if (jlo.lt.1) then
				jlo=0
			else if (x.lt.xx(jlo).eqv.ascnd) then
				jhi=jlo
				inc=inc+inc
				goto 2
			endif
		endif
3 		if (jhi-jlo.eq.1) then
			if(x.eq.xx(n)) jlo=n-1
			if(x.eq.xx(1)) jlo=1
			return
		endif
		jm = (jhi+jlo)/2
		if (x.ge.xx(jm).eqv.ascnd) then
			jlo=jm
		else
			jhi=jm
		endif
		goto 3
	end subroutine hunt

! ---------------------------------------------------------
! Given xa(:) and ya(:) which tabulate a function, returns the interpolation using
! linear interpolation of vector x(:) in y(:)
! ---------------------------------------------------------
	subroutine lin_interpol(xa,ya,x,y,zero_outside)
   real(kind=8), INTENT(IN) :: xa(:), ya(:), x(:)
   real(kind=8), INTENT(INOUT) :: y(:)
   integer :: i, n, na
   integer :: loc, jlo
   logical :: zero_outside
   real(kind=8) :: factor
   

   	n = size(x)
   	na = size(xa)

! If zero_outside is set, put to zero everything outside the intervals where
! data is available
   	if (zero_outside) then
   		factor = 0.d0
   	else
   		factor = 1.d0
   	endif

      do i = 1, n
      	call hunt(xa,na,x(i),jlo)
         if (jlo == 0) then
         	y(i) = ya(1) * factor
         else if (jlo == na) then
         	y(i) = ya(na) * factor
         else
            y(i) = (ya(jlo+1)-ya(jlo))/(xa(jlo+1)-xa(jlo)) * (x(i)-xa(jlo)) + ya(jlo)
         endif
      enddo

   end subroutine lin_interpol

! ---------------------------------------------------------
! This routine performs a parabolic interpolation of the parabola that
! passes through 3 points x,y to a vector xx,yy
! ---------------------------------------------------------
	subroutine threept_parab_interp(nn, x, y, xx, yy)  
  	integer :: nn
  	real(kind=4), dimension (3) :: x, y
  	real(kind=4), dimension (nn) :: xx, yy
  	real(kind=4) :: den, a, b, c, x1, x2, x3, y1, y2, y3

  		x1=x(1)
  		x2=x(2)
		x3=x(3)
		y1=y(1)
		y2=y(2)
		y3=y(3)
		den = (x1 - x2)*(x1 - x3)*(x2 - x3)
		a = (x3*(y2-y1)+x2*(y1-y3)+x1*(y3-y2))/den
		b = (x3*x3*(y1-y2)+x1*x1*(y2-y3)+x2*x2*(y3-y1))/den
		c = (x1*x3*y2*(x3-x1)+x2*x2*(x3*y1-x1*y3)+ &
				x2*(x1*x1*y3-x3*x3*y1))/den
		yy = a*xx*xx + b*xx + c
	end subroutine threept_parab_interp

! ---------------------------------------------------------
! This routine performs an interpolation of the curve that
! passes through x,y to xx,yy. The interpolation is performed joining
! adjacent points by a straight line and then smoothing the result 
! by convolving it with a Gaussian whose width is a half of the distance
! between the two closest nodes.
! ---------------------------------------------------------
	subroutine smoothed_lines(n, nn, x, y, xx, yy)
	integer :: n, nn, inode, ipoint, jpoint, i
	real(kind=4), dimension (n) :: x, y
	real(kind=4), dimension (nn) :: xx, yy, yy2
	real(kind=8), dimension (n) :: xdouble, ydouble
	real(kind=8), dimension (nn) :: xxdouble, yydouble
	real :: x1, x2, y1, y2, closest, expo, norm

		xdouble = x
		ydouble = y
		xxdouble = xx
		call spline(xdouble,ydouble,xxdouble,yydouble)
		yy = real(yydouble)

		return

! compute segments
	yy=0.
	do inode=1, n-1
		x1=x(inode)
		x2=x(inode+1)
		y1=y(inode)
		y2=y(inode+1)
		do ipoint=1, nn
			if (xx(ipoint) .ge. x1 .and. xx(ipoint) .le. x2) then
				yy(ipoint)=y1+(y2-y1)/(x2-x1)*(xx(ipoint)-x1)
			endif
		enddo
	enddo

! Smooth the result
! Find the closest nodes

	closest=abs(x(1)-x(2))
	do inode=2, n-1
		if (abs(x(inode)-x(inode+1)) .lt. closest) &
				closest=abs(x(inode)-x(inode+1))
	end do

! convolve (note that the model is not necessarily equispaced)
	closest=closest/2.
	yy2=0.
	do ipoint=1, nn
		norm=0.
		do jpoint=1,nn
			x1=xx(jpoint)-xx(ipoint)
			if (x1*x1/closest/closest .lt. 25.) then
				expo=exp(-x1*x1/(closest*closest))
				yy2(ipoint)=yy2(ipoint)+yy(jpoint)*expo
				norm=norm+expo
			end if
		end do
		yy2(ipoint)=yy2(ipoint)/norm
	end do
	yy=yy2
		
	end subroutine smoothed_lines

!-------------------------------------------------------------------
! Add constant/linear/polynomial perturbation to the atmosphere
!-------------------------------------------------------------------
	subroutine add_perturbation(n, ltau500, original, perturbed, pos_nodes, values_nodes)
	integer :: n, pos_nodes(n)
	real(kind=4), dimension(:) :: ltau500, original, perturbed
	real(kind=8) :: values_nodes(n), slope
	integer :: i, j

		select case(n)
			case(1)
				perturbed = original + values_nodes(1)
			case(2)
				slope = (values_nodes(2) - values_nodes(1)) / (ltau500(pos_nodes(2))- ltau500(pos_nodes(1)))
				perturbed = original + values_nodes(1) + slope * (ltau500 - ltau500(pos_nodes(1)))
			case(3)
				call threept_parab_interp(Params%n_points, ltau500(pos_nodes(1:n)), real(values_nodes(1:n)), ltau500, perturbed)
				perturbed = original + perturbed
			case(4:)
				call smoothed_lines(n, Params%n_points, ltau500(pos_nodes(1:n)), real(values_nodes(1:n)), ltau500, perturbed)
				perturbed = original + perturbed
		end select

	end subroutine add_perturbation


!-------------------------------------------------------------------
! Modify all components to the new values at the nodes
!-------------------------------------------------------------------
	subroutine set_modified_atmospheres(last)
	real(kind=8) :: last(:)
	integer :: i, j, k, loop

! First accumulate the perturbations for each node, variable and component
		loop = 1
		do j = 1, n_param_per_component
			do i = 1, atmos%ncomponents
				do k = 1, atmos%n_nodes(i,j)
					atmos%value_nodes(i,j,k) = last(loop)
					loop = loop + 1
				enddo
			enddo
		enddo

! Then do the actual perturbation
		loop = 1
		do j = 1, n_param_per_component
			do i = 1, atmos%ncomponents
				select case(j)
					case(1)
						call add_perturbation(atmos%n_nodes(i,j), atmos%stratification(i)%ltau_500, atmos%stratification(i)%temp, &
							atmos%perturbed(i)%temp, atmos%pos_nodes(i,j,1:atmos%n_nodes(i,j)), atmos%value_nodes(i,j,1:atmos%n_nodes(i,j)))
					case(2)
						call add_perturbation(atmos%n_nodes(i,j), atmos%stratification(i)%ltau_500, atmos%stratification(i)%v_mic, &
							atmos%perturbed(i)%v_mic, atmos%pos_nodes(i,j,1:atmos%n_nodes(i,j)), atmos%value_nodes(i,j,1:atmos%n_nodes(i,j)))
					case(3)
						call add_perturbation(atmos%n_nodes(i,j), atmos%stratification(i)%ltau_500, atmos%stratification(i)%b_str, &
							atmos%perturbed(i)%b_str, atmos%pos_nodes(i,j,1:atmos%n_nodes(i,j)), atmos%value_nodes(i,j,1:atmos%n_nodes(i,j)))
					case(4)
						call add_perturbation(atmos%n_nodes(i,j), atmos%stratification(i)%ltau_500, atmos%stratification(i)%b_inc, &
							atmos%perturbed(i)%b_inc, atmos%pos_nodes(i,j,1:atmos%n_nodes(i,j)), atmos%value_nodes(i,j,1:atmos%n_nodes(i,j)))
					case(5)
						call add_perturbation(atmos%n_nodes(i,j), atmos%stratification(i)%ltau_500, atmos%stratification(i)%b_azi, &
							atmos%perturbed(i)%b_azi, atmos%pos_nodes(i,j,1:atmos%n_nodes(i,j)), atmos%value_nodes(i,j,1:atmos%n_nodes(i,j)))
					case(6)
						call add_perturbation(atmos%n_nodes(i,j), atmos%stratification(i)%ltau_500, atmos%stratification(i)%v_los, &
							atmos%perturbed(i)%v_los, atmos%pos_nodes(i,j,1:atmos%n_nodes(i,j)), atmos%value_nodes(i,j,1:atmos%n_nodes(i,j)))
					case(7)
 						atmos%ff(i) = atmos%value_nodes(i,j,1)
				end select
				loop = loop + 1
			enddo
		enddo

	end subroutine set_modified_atmospheres


end module maths_mod
